@@ -19,6 +19,7 @@ except ModuleNotFoundError:
 import io
 import json
 from base64 import b64encode
+from threading import Lock
 
 try:
     from collections.abc import Iterable
@@ -67,6 +68,10 @@ cursors_str = {
     cursors.WAIT: 'wait',
 }
 
+# threading.Lock to prevent multiple threads from accessing globals such as Gcf
+_lock = Lock()
+
+
 
 def connection_info():
     """
@@ -75,18 +80,19 @@ def connection_info():
     use.
 
     """
-    result = []
-    for manager in Gcf.get_all_fig_managers():
-        fig = manager.canvas.figure
-        result.append(
-            '{} - {}'.format(
-                (fig.get_label() or f"Figure {manager.num}"),
-                manager.web_sockets,
+    with _lock:
+        result = []
+        for manager in Gcf.get_all_fig_managers():
+            fig = manager.canvas.figure
+            result.append(
+                '{} - {}'.format(
+                    (fig.get_label() or f"Figure {manager.num}"),
+                    manager.web_sockets,
+                )
             )
-        )
-    if not is_interactive():
-        result.append(f'Figures pending show: {len(Gcf._activeQue)}')
-    return '\n'.join(result)
+        if not is_interactive():
+            result.append(f'Figures pending show: {len(Gcf._activeQue)}')
+        return '\n'.join(result)
 
 
 class Toolbar(DOMWidget, NavigationToolbar2WebAgg):
@@ -122,6 +128,10 @@ class Toolbar(DOMWidget, NavigationToolbar2WebAgg):
 
         self.on_msg(self.canvas._handle_message)
 
+    def save_figure(self, *args):
+        """Override to use rcParams-aware save."""
+        self.canvas._send_save_buffer()
+
     def export(self):
         buf = io.BytesIO()
         self.canvas.figure.savefig(buf, format='png', dpi='figure')
@@ -134,7 +144,8 @@ class Toolbar(DOMWidget, NavigationToolbar2WebAgg):
             width = pwidth / self.canvas._dpi_ratio
         data = "<img src='data:image/png;base64,{0}' width={1}/>"
         data = data.format(b64encode(buf.getvalue()).decode('utf-8'), width)
-        display(HTML(data))
+        with _lock:
+            display(HTML(data))
 
     @default('toolitems')
     def _default_toolitems(self):
@@ -320,6 +331,52 @@ class Canvas(DOMWidget, FigureCanvasWebAggCore):
         # Actually send the data
         self.send({'data': '{"type": "binary"}'}, buffers=[data])
 
+    def download(self):
+        """
+        Trigger a download of the figure respecting savefig rcParams.
+
+        This is a programmatic way to trigger the same download that happens
+        when the user clicks the Download button in the toolbar.
+
+        The figure will be saved using all applicable savefig.* rcParams
+        including format, dpi, transparent, facecolor, etc.
+
+        Examples
+        --------
+        >>> fig, ax = plt.subplots()
+        >>> ax.plot([1, 2, 3], [1, 4, 2])
+        >>> fig.canvas.download()  # Downloads with current rcParams
+
+        >>> # Download as PDF
+        >>> plt.rcParams['savefig.format'] = 'pdf'
+        >>> fig.canvas.download()
+
+        >>> # Download with custom DPI
+        >>> plt.rcParams['savefig.dpi'] = 300
+        >>> fig.canvas.download()
+        """
+        self._send_save_buffer()
+
+    def _send_save_buffer(self):
+        """Generate figure buffer respecting savefig rcParams and send to frontend."""
+        buf = io.BytesIO()
+
+        # Call savefig WITHOUT any parameters - fully respects all rcParams
+        self.figure.savefig(buf)
+
+        # Get the format that was used
+        fmt = rcParams.get('savefig.format', 'png')
+
+        # Get the buffer data
+        data = buf.getvalue()
+
+        # Send to frontend with format metadata
+        msg_data = {
+            "type": "save",
+            "format": fmt
+        }
+        self.send({'data': json.dumps(msg_data)}, buffers=[data])
+
     def new_timer(self, *args, **kwargs):
         return TimerTornado(*args, **kwargs)
 
@@ -384,74 +441,9 @@ class Canvas(DOMWidget, FigureCanvasWebAggCore):
         data = self._repr_mimebundle_(**kwargs)
         display(data, raw=True)
 
-    if matplotlib.__version__ < '3.4':
-        # backport the Python side changes to match the js changes
-        def _handle_key(self, event):
-            _SPECIAL_KEYS_LUT = {
-                'Alt': 'alt',
-                'AltGraph': 'alt',
-                'CapsLock': 'caps_lock',
-                'Control': 'control',
-                'Meta': 'meta',
-                'NumLock': 'num_lock',
-                'ScrollLock': 'scroll_lock',
-                'Shift': 'shift',
-                'Super': 'super',
-                'Enter': 'enter',
-                'Tab': 'tab',
-                'ArrowDown': 'down',
-                'ArrowLeft': 'left',
-                'ArrowRight': 'right',
-                'ArrowUp': 'up',
-                'End': 'end',
-                'Home': 'home',
-                'PageDown': 'pagedown',
-                'PageUp': 'pageup',
-                'Backspace': 'backspace',
-                'Delete': 'delete',
-                'Insert': 'insert',
-                'Escape': 'escape',
-                'Pause': 'pause',
-                'Select': 'select',
-                'Dead': 'dead',
-                'F1': 'f1',
-                'F2': 'f2',
-                'F3': 'f3',
-                'F4': 'f4',
-                'F5': 'f5',
-                'F6': 'f6',
-                'F7': 'f7',
-                'F8': 'f8',
-                'F9': 'f9',
-                'F10': 'f10',
-                'F11': 'f11',
-                'F12': 'f12',
-            }
-
-            def handle_key(key):
-                """Handle key values"""
-                value = key[key.index('k') + 1 :]
-                if 'shift+' in key:
-                    if len(value) == 1:
-                        key = key.replace('shift+', '')
-                if value in _SPECIAL_KEYS_LUT:
-                    value = _SPECIAL_KEYS_LUT[value]
-                key = key[: key.index('k')] + value
-                return key
-
-            key = handle_key(event['key'])
-            e_type = event['type']
-            guiEvent = event.get('guiEvent', None)
-            if e_type == 'key_press':
-                self.key_press_event(key, guiEvent=guiEvent)
-            elif e_type == 'key_release':
-                self.key_release_event(key, guiEvent=guiEvent)
-
-        handle_key_press = handle_key_release = _handle_key
-
 
 class FigureManager(FigureManagerWebAgg):
-    if matplotlib.__version__ < "3.6":
+    if matplotlib.__version_info__ < (3, 6):
         ToolbarCls = Toolbar
 
     def __init__(self, canvas, num):
@@ -462,7 +454,8 @@ class FigureManager(FigureManagerWebAgg):
     def show(self):
         if self.canvas._closed:
             self.canvas._closed = False
-            display(self.canvas)
+            with _lock:
+                display(self.canvas)
         else:
             self.canvas.draw_idle()
 
@@ -480,83 +473,86 @@ class _Backend_ipympl(_Backend):
 
     @staticmethod
     def new_figure_manager_given_figure(num, figure):
-        canvas = Canvas(figure)
-        if 'nbagg.transparent' in rcParams and rcParams['nbagg.transparent']:
-            figure.patch.set_alpha(0)
-        manager = FigureManager(canvas, num)
+        with _lock:
+            canvas = Canvas(figure)
+            if 'nbagg.transparent' in rcParams and rcParams['nbagg.transparent']:
+                figure.patch.set_alpha(0)
+            manager = FigureManager(canvas, num)
 
-        if is_interactive():
-            _Backend_ipympl._to_show.append(figure)
-            figure.canvas.draw_idle()
+            if is_interactive():
+                _Backend_ipympl._to_show.append(figure)
+                figure.canvas.draw_idle()
 
-        def destroy(event):
-            canvas.mpl_disconnect(cid)
-            Gcf.destroy(manager)
+            def destroy(event):
+                canvas.mpl_disconnect(cid)
+                Gcf.destroy(manager)
 
-        cid = canvas.mpl_connect('close_event', destroy)
+            cid = canvas.mpl_connect('close_event', destroy)
 
-        # Only register figure for showing when in interactive mode (otherwise
-        # we'll generate duplicate plots, since a user who set ioff() manually
-        # expects to make separate draw/show calls).
-        if is_interactive():
-            # ensure current figure will be drawn.
-            try:
-                _Backend_ipympl._to_show.remove(figure)
-            except ValueError:
-                # ensure it only appears in the draw list once
-                pass
-            # Queue up the figure for drawing in next show() call
-            _Backend_ipympl._to_show.append(figure)
-            _Backend_ipympl._draw_called = True
+            # Only register figure for showing when in interactive mode (otherwise
+            # we'll generate duplicate plots, since a user who set ioff() manually
+            # expects to make separate draw/show calls).
+            if is_interactive():
+                # ensure current figure will be drawn.
+                try:
+                    _Backend_ipympl._to_show.remove(figure)
+                except ValueError:
+                    # ensure it only appears in the draw list once
+                    pass
+                # Queue up the figure for drawing in next show() call
+                _Backend_ipympl._to_show.append(figure)
+                _Backend_ipympl._draw_called = True
 
-        return manager
+            return manager
 
     @staticmethod
     def show(block=None):
-        # # TODO: something to do when keyword block==False ?
-        interactive = is_interactive()
+        with _lock:
+            # # TODO: something to do when keyword block==False ?
+            interactive = is_interactive()
 
-        manager = Gcf.get_active()
-        if manager is None:
-            return
+            manager = Gcf.get_active()
+            if manager is None:
+                return
 
-        try:
-            display(manager.canvas)
-            # metadata=_fetch_figure_metadata(manager.canvas.figure)
+            try:
+                display(manager.canvas)
+                # metadata=_fetch_figure_metadata(manager.canvas.figure)
 
-            # plt.figure adds an event which makes the figure in focus the
-            # active one. Disable this behaviour, as it results in
-            # figures being put as the active figure after they have been
-            # shown, even in non-interactive mode.
-            if hasattr(manager, '_cidgcf'):
-                manager.canvas.mpl_disconnect(manager._cidgcf)
+                # plt.figure adds an event which makes the figure in focus the
+                # active one. Disable this behaviour, as it results in
+                # figures being put as the active figure after they have been
+                # shown, even in non-interactive mode.
+                if hasattr(manager, '_cidgcf'):
+                    manager.canvas.mpl_disconnect(manager._cidgcf)
 
-            if not interactive:
-                Gcf.figs.pop(manager.num, None)
-        finally:
-            if manager.canvas.figure in _Backend_ipympl._to_show:
-                _Backend_ipympl._to_show.remove(manager.canvas.figure)
+                if not interactive:
+                    Gcf.figs.pop(manager.num, None)
+            finally:
+                if manager.canvas.figure in _Backend_ipympl._to_show:
+                    _Backend_ipympl._to_show.remove(manager.canvas.figure)
 
 
 def flush_figures():
-    backend = matplotlib.get_backend()
-    if backend in ('widget', 'ipympl', 'module://ipympl.backend_nbagg'):
-        if not _Backend_ipympl._draw_called:
-            return
+    with _lock:
+        backend = matplotlib.get_backend()
+        if backend in ('widget', 'ipympl', 'module://ipympl.backend_nbagg'):
+            if not _Backend_ipympl._draw_called:
+                return
 
-        try:
-            # exclude any figures that were closed:
-            active = {fm.canvas.figure for fm in Gcf.get_all_fig_managers()}
-
-            for fig in [fig for fig in _Backend_ipympl._to_show if fig in active]:
-                # display(fig.canvas, metadata=_fetch_figure_metadata(fig))
-                display(fig.canvas)
-        finally:
-            # clear flags for next round
-            _Backend_ipympl._to_show = []
-            _Backend_ipympl._draw_called = False
+            try:
+                # exclude any figures that were closed:
+                active = {fm.canvas.figure for fm in Gcf.get_all_fig_managers()}
+                for fig in [fig for fig in _Backend_ipympl._to_show if fig in active]:
+                    # display(fig.canvas, metadata=_fetch_figure_metadata(fig))
+                    display(fig.canvas)
+            finally:
+                # clear flags for next round
+                _Backend_ipympl._to_show = []
+                _Backend_ipympl._draw_called = False
 
 
-ip = get_ipython()
-if ip is not None:
-    ip.events.register('post_execute', flush_figures)
+with _lock:
+    ip = get_ipython()
+    if ip is not None:
+        ip.events.register('post_execute', flush_figures)
